/*
   A BLE turbo trainer controller for Zwift

   Program receives environment date from the Zwift game (via BLE) and adjusts the resistance of a turbo trainer using a stepper motor
   Uses an ESP32 and includes an LCD display for feedback.

   Copyright 2020 Peter Everett
   v1.0 Aug 2020 - Initial version

   This work is licensed the GNU General Public License v3 (GPL-3)
   Modified for Proform TDF bike generation 2 by Greg Masik April 1, 2021. Removed stepper motor and LCD display
   April 28, modified to allow for zwift half grade for negative grades. May 10, 2021 added I2C LCD display for grade.
   January 5, 2022- Added case 0x01 to allow for reset request from Zwift after Dec 2021 update.

*/

#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>
#include <LCD_I2C.h>

LCD_I2C lcd(0x27); // Default address of most PCF8574 modules, change according

const int upperInclineClamp = 10;
const int lowerInclineClamp = -5; //changed from -10

const int upPin = 26;
const int downPin = 17;

float gradeFloat = 0.0;
int_least16_t grade = 0;

int roundedGrade = 0;
int currentBikeGrade = 0;

BLEServer* pServer = NULL;
BLECharacteristic* pIndoorBike = NULL;
BLECharacteristic* pFeature = NULL;
BLECharacteristic* pControlPoint = NULL;
BLECharacteristic* pStatus = NULL;

BLEAdvertisementData advert;
BLEAdvertisementData scan_response;
BLEAdvertising *pAdvertising;

bool deviceConnected = false;
bool oldDeviceConnected = false;
int value = 0;                   //This is the value sent as "nothing".  We need to send something for some of the charactistics or it won't work.

#define FTMSDEVICE_FTMS_UUID "00001826-0000-1000-8000-00805F9B34FB"
#define FTMSDEVICE_INDOOR_BIKE_CHAR_UUID "00002AD2-0000-1000-8000-00805F9B34FB"
//#define FTMSDEVICE_RESISTANCE_RANGE_CHAR_UUID "00002AD6-0000-1000-8000-00805F9B34FB"
//#define FTMSDEVICE_POWER_RANGE_CHAR_UUID "00002AD8-0000-1000-8000-00805F9B34FB"
#define FTMSDEVICE_FTMS_FEATURE_CHAR_UUID "00002ACC-0000-1000-8000-00805F9B34FB"
#define FTMSDEVICE_FTMS_CONTROL_POINT_CHAR_UUID "00002AD9-0000-1000-8000-00805F9B34FB"
#define FTMSDEVICE_FTMS_STATUS_CHAR_UUID "00002ADA-0000-1000-8000-00805F9B34FB"

//response/acknowledgement to send to the client after writing to control point
uint8_t replyDs[3] = {0x80, 0x00, 0x01};   // set up replyDS array with 3 byte values for control point resp to zwift

class MyServerCallbacks: public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
    };

    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
    }
};

void setup() {


  pinMode(upPin, OUTPUT);
  pinMode(downPin, OUTPUT);
  digitalWrite(upPin, LOW);
  digitalWrite(downPin, LOW);

  lcd.begin(); // If you are using more I2C devices using the Wire library use lcd.begin(false)
  // this stop the library(LCD_I2C) from calling Wire.begin()
  lcd.backlight(); //turn on backlight

  //Serial.begin(115200);
  //Serial.println("Starting initialisation routine");

  //Setup BLE
  //Serial.println("Creating BLE server...");
  BLEDevice::init("ESP32");

  // Create the BLE Server
  pServer = BLEDevice::createServer();
  pServer->setCallbacks(new MyServerCallbacks());

  // Create the BLE Service
  // Serial.println("Define service...");
  BLEService *pService = pServer->createService(FTMSDEVICE_FTMS_UUID);

  // Create BLE Characteristics
  //Serial.println("Define characteristics");
  pIndoorBike = pService->createCharacteristic(FTMSDEVICE_INDOOR_BIKE_CHAR_UUID, BLECharacteristic::PROPERTY_NOTIFY);
  pIndoorBike->addDescriptor(new BLE2902());
  pControlPoint = pService->createCharacteristic(FTMSDEVICE_FTMS_CONTROL_POINT_CHAR_UUID, BLECharacteristic::PROPERTY_INDICATE | BLECharacteristic::PROPERTY_WRITE);
  pControlPoint->addDescriptor(new BLE2902());
  pFeature = pService->createCharacteristic(FTMSDEVICE_FTMS_FEATURE_CHAR_UUID, BLECharacteristic::PROPERTY_READ);
  pFeature->addDescriptor(new BLE2902());
  pStatus = pService->createCharacteristic(FTMSDEVICE_FTMS_STATUS_CHAR_UUID, BLECharacteristic::PROPERTY_NOTIFY);
  pStatus->addDescriptor(new BLE2902());

  // Start the service
  //Serial.println("Staring BLE service...");
  pService->start();

  // Start advertising
  //Serial.println("Define the advertiser...");
  pAdvertising = BLEDevice::getAdvertising();

  pAdvertising->setScanResponse(true);
  pAdvertising->addServiceUUID(FTMSDEVICE_FTMS_UUID);
  pAdvertising->setMinPreferred(0x06);  // set value to 0x00 to not advertise this parameter
  //Serial.println("Starting advertiser...");
  BLEDevice::startAdvertising();
  //Serial.println("Waiting a client connection to notify...");

  //Serial.println("Waiting for");

  // Serial.println("connection...");

}

/*
  The main loop which runs continuously.
  Checks whether something is connected
  - Receive bike simulation data, adjust the trainer and send acknowledgement back
*/


void loop() {
  char windStr[6];  //6 spot array 1 byte each -wind speed
  char gradeStr[6]; //6 spot array 1 byte each - grade
  char rStr[4];  //4spot array 1 byte each  - rolling res
  char wStr[4];  //4 spot array 1 byte each - wind coef
  // notify changed value
  if (deviceConnected) {
    //indoor bike char values
    uint8_t bikeData[19] = {};  //set all the bytes to zero .  If we're only transmitting power, there's no point transmitting beyond byte 18
    bikeData[0] = 0x01;   // flags for Indoor bike-set bit 6 low of byte 0 to say power data not present
    //bit 1 set high for inst speed NOT present
    // bikeData[15] = 0x00;  //set bytes? 15 and 16 for the power (note that this is a signed int16 no power (0)
    // bikeData[16] = 0x00;
    uint8_t flags = bikeData[0]; //set to 0x01
    uint8_t dataToSend[1] = {flags};
    pIndoorBike->setValue(dataToSend, 1);
    pIndoorBike->notify(); //notify zwift

    //configure machine feature characteristic 8 fields- only set indoor bike simulation bit(field 5)
    uint8_t feature[8] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00};   // 2^13 = bike simulation bit set byte 5.
    pFeature->setValue(feature, 8); //send data to feature char- 8 locations from feature array
    pStatus->setValue(value);  //send data to status char- value =0
    pStatus->notify(); //notify zwift

    //Get the data written to the control point
    std::string rxValue = pControlPoint->getValue();
    if (rxValue.length() == 0) {
      // Serial.println("No data received...");
    } else {

      switch (rxValue[0]) {
        case 0x00:  //Request control from zwift
          //reply with 0x80, 0x00, 0x01 to say OK
          replyDs[1] = rxValue[0];
          pControlPoint->setValue(replyDs, 3);
          pControlPoint->indicate();
          break;

        case 0x01: //reset
          //reply with 0x80, 0x01, 0x01 to say OK          
          replyDs[1] = rxValue[0];
          pControlPoint->setValue(replyDs, 3);
          pControlPoint->indicate();
          break;
          
        case 0x07:  //Start/resume
          //reply with 0x80, 0x07, 0x01 to say OK
          replyDs[1] = rxValue[0];
          pControlPoint->setValue(replyDs, 3);
          pControlPoint->indicate();
          break;
        case 0x11:  //receive simulation parameters
          /*
            In the case of Zwift:
            Wind is always constant, even with an aero boost (you just go faster)
            When trainer difficulty is set to maximum, the grade is as per the BLE spec (500 = 5%)
            Rolling resistance only ever seems to be 0 or 255.  255 occurs on mud/off road sections but not always (short sections only, like a patch of sand?)
            Wind coefficient doesn't change during a ride, even with an aero boost.  It may change depending on bike used in the game.
            Note: only using grade for TDF bike in 1% increments*/

          int16_t wind = rxValue[2] << 8; //register changed from [1] in turbotrainer original
          wind |= rxValue[1]; //register changed from [0]in turbotrainer original

          uint8_t Rres = rxValue[5]; //register changed from [4]in turbotrainer original
          uint8_t Wres = rxValue[6]; //register changed from [5]in turbotrainer original

          grade = rxValue[4] << 8;
          grade |= rxValue[3];

          gradeFloat = grade / 100.0;

          if (gradeFloat < 0)
          {
            gradeFloat = (gradeFloat * 2);
          }


          lcd.setCursor(0, 0);
          lcd.print("Game Grade= ");
          lcd.print(gradeFloat);




          // Serial.print("gradeFloat= ");
          // Serial.println(gradeFloat);
          //Serial.print("wind speed= ");
          //Serial.println(wind);
          //Serial.print("Rolling res= ");
          //Serial.println(Rres);
          //Serial.print("Wind Res= ");
          //Serial.println(Wres);



          roundedGrade = round(gradeFloat);



          //  Serial.print("rounded grade = ");
          // Serial.println(roundedGrade);

          if (roundedGrade >= upperInclineClamp)
          {
            roundedGrade = upperInclineClamp;
          }
          if (roundedGrade <= lowerInclineClamp)
          {
            roundedGrade = lowerInclineClamp;
          }
          if (roundedGrade == currentBikeGrade)
          {
            delay(275); //provides a delay for  the no change condition- same as other for uniformity
          }

          if (roundedGrade > currentBikeGrade)
          {
            currentBikeGrade = (currentBikeGrade + 1);
            digitalWrite(upPin, HIGH);
            delay(275);  //reduce delay from 500 as getting double strikes inc from 250 as missing strikes
            digitalWrite(upPin, LOW);
          }
          if (roundedGrade < currentBikeGrade)
          {
            currentBikeGrade = (currentBikeGrade - 1);
            digitalWrite(downPin, HIGH);
            delay(275); //from 250
            digitalWrite(downPin, LOW);
          }


          //short delay to give the motor a break if the grade is changing rapidly
          // Serial.print("Current Bike Grade = ");
          //Serial.println(currentBikeGrade);

          float fltCurBikeGrade = currentBikeGrade;
          lcd.setCursor(0, 1);
          lcd.print("Bike Grade= ");
          lcd.print(fltCurBikeGrade);
          //lcd.println();

          delay(1000);
          // Serial.println("One Second Delay");
          break;
      }
    }
    delay(50); // Delay to stop bluetooth stack getting congested-added to slow down incline changes
    // changed from 50
  }
  // disconnecting
  if (!deviceConnected && oldDeviceConnected) {
    delay(300); // give the bluetooth stack the chance to get things ready
    pServer->startAdvertising(); // restart advertising
    //Serial.println("Nothing connected, start advertising");
    oldDeviceConnected = deviceConnected;

  }
  // connecting
  if (deviceConnected && !oldDeviceConnected) {
    // do stuff here on connecting
    oldDeviceConnected = deviceConnected;
  }
}
